#include <AudioToolbox/AudioToolbox.h>
#include <CoreAudio/CoreAudio.h>
#include "cahelper.h"

#include <vector>
#include "soundfileplayer.h"

#include "Complex.h"
#include "Compressor.h"
#include "Control.h"
#include "SineMod.h"

class Overdrive : public Processor {
    float drive;
public:
    Overdrive(float d) {
        drive = d;
    }
    
    float Process(float input) {
        return tanh(input * drive);
    }
};

class Filter : public Processor {
public:
    Complex pole;
    Complex zero;
    Complex previousInput;
    Complex previousOutput;
    
    // rakentaja; anetaan napa ja nolla, oletusarvoisesti origossa
    Filter(Complex p = 0, Complex z = 0) {
        pole = p;
        zero = z;
        previousInput = 0;
        previousOutput = 0;
    }
    
    // lasketaan suotimen magnitudivaste taajuudella [0, Pi]
    float MagnitudeAt(float frequency) {
        Complex z = Complex::Polar(frequency, 1);
        // |H(z)| = |(z - zero) / (z - pole)|
        return ((z-zero)/(z-pole)).Absolute();
    }
    
    float Process(float input) {
        // y[t] = x[t] + zero * x[t-1] - pole * y[t-1]
        Complex output = input + zero * previousInput - pole * previousOutput;
        
        previousOutput = output;
        previousInput  = input;
        
        return output.real;
    }
};

class Echo : public Processor {
    int delayTime;
    float feedback;
    std::vector<float> delayLine;
    int delayPos;
    float mixWetDry;
public:
    Echo(float mix, int delay, float fb) {
        delayTime = delay;
        feedback = fb;
        delayPos = 0;
        mixWetDry = mix;
        delayLine.resize(delayTime);
    }
    
    float Process(float input) {
        float output = delayLine[delayPos];
        delayLine[delayPos] = input + feedback * output;
        delayPos = delayPos + 1;
        if (delayPos >= delayLine.size( )) delayPos = 0;
        return input + mixWetDry * (output - input); // LERP, linear interpolation
        // input * (1 - mixWetDry) + output * mixWetDry
    }
};

class Reverb : public Processor {
    std::vector<Echo> combs;
    float wetDryMix;
public:
    // tämä kaava laskee RT60-ajan toteuttavan takaisinkytkentäkertoimen
    // halutunpituiselle viivelinjalle
    float GetFeedback(float rt60, int delay) {
        // 0.001 vastaa -60dB vaimenemista
        // eksponentti johdetaan siitä, kuinka monta kertaa signaali kulkee
        // takaisinkytkentäkertoimen läpi ajassa, jossa sen tulee vaimentua
        // -60dB
        return pow(0.001, (float)delay / (PlayerSampleRate * rt60) );
    }
    
    Reverb(float mix, float rt60) {
        // lisätään kaikuun 8 viivelinjaa
        for(int i=0;i<8;i++) {
            // viivelinjan pituus on 900 * 2^(i/8), eli eksponentiaalinen jakauma
            // välillä 900-1800
            int delayTime = 900 * pow(2, i / 8.0);
            // luodaan viivelinja ja lisätään vektoriin
            combs.push_back(
                            Echo(1, delayTime, GetFeedback(rt60, delayTime) ));
        }
        wetDryMix = mix;
    }
    
    float Process(float input) {
        float verb = 0;
        // syötetään input kaikkiin viivelinjoihin ja summataan niiden output
        for(Echo& e : combs) verb = verb + e.Process(input);
        // lasketaan kokonaisoutputiin lisäksi kuiva signaali
        return input + wetDryMix * (verb * 0.1 - input);
    }
};

int main (int argc, const char * argv[]) {
    SineMod myOsc(440);
    float f = 440;
    
    Control myController(8, [&]() {
        f = f + 40;
        myOsc.SetFrequency(f);
    });
    
    PlayAudio( myController >> myOsc >> SoundfileRecorder("/Users/CMTUser/Documents/test.wav") , 5.0);
    return 0;
}























